--[[--------------------------------------------------------------------------
  Developed by Nathan Pieper - nrpieper (@) gmail (dot) com
  Misspelled is an interactive chat text spell checker for World of Warcraft

  This code freely distributed for your use in any GPL compliant project.
  Portions of this code are covered by the Gnu Plublic License (GPL)
  Dictionaries use the OpenOffice/HunSpell/ASpell dictionary format
  (http://wiki.services.openoffice.org/wiki/Dictionaries)
  Phonetic cache information created via the NetSpell dictionary util.
  References & Credits:
  NetSpell: http://www.loresoft.com/projects/netspell
  Metaphone Algorithm http://aspell.net/metaphone/
  Edit Distance Algorithm http://en.wikipedia.org/wiki/Levenshtein_distance
  Portions Copyright (c) 2003, Paul Welter All rights reserved.

--------------------------------------------------------------------------]]--

Misspelled = {
	VERSION = "Misspelled 0.1.06",
	queue = {},
	queuePos = 0,
}

local L = MisspelledLocals

local SPELLED_WRONG_HIGHLIGHT = "|cffff91c8"
local WordCache = {}
local WordLocations = {}
local RightClickedWord = nil
local RightClickedWordStartPos
local RightClickedWordStartPos


function Misspelled_SendChatMessage(message, type, language, channel)
	--Before message is sent, remove any highlighting
	--The Wow client will disconnect if you attempt to send a colored chat message.
	local newText = message
	newText = string.gsub(newText, "|cffff91c8", "")
	newText = string.gsub(newText, "|r", "")

	Misspelled_Orig_SendChatMessage(newText, type, language, channel);
end

--Split a test script, at patt delimiter, into a table
function Misspelled:split(str, patt)
	vals = {}; valindex = 0; word = ""
	-- need to add a trailing separator to catch the last value.
	str = str .. patt
	for i = 1, string.len(str) do

		cha = string.sub(str, i, i)
		if cha ~= patt then
			word = word .. cha
		else
			if word ~= nil then
				vals[valindex] = word
				valindex = valindex + 1
				word = ""
			else
				-- in case we get a line with no data.
				break
			end
		end

	end
	return vals
end

function Misspelled:tprint (t, indent, done)
  -- in case we run it standalone
  local Note = Note or print
--  local Tell = Tell or io.write

  -- show strings differently to distinguish them from numbers
  local function show (val)
    if type (val) == "string" then
      return '"' .. val .. '"'
    else
      return tostring (val)
    end -- if
  end -- show
  -- entry point here
  done = done or {}
  indent = indent or 0
  for key, value in pairs (t) do
    print(string.rep (" ", indent)) -- indent it
    if type (value) == "table" and not done [value] then
      done [value] = true
      Note (show (key), ":");
      tprint (value, indent + 2, done)
    else
      print(show (key), "=")
      print (show (value))
    end
  end
end




function Misspelled:Initialize(frame)
	--Misspelled:OnSizeChanged(frame)
	--table.insert(UISpecialFrames, "MisspelledFrame")
	--PlaySound("igMainMenuOpen");
	--self:UpdateButtons()
end


function Misspelled_OnLoad()
	this:RegisterEvent("VARIABLES_LOADED")
	this:RegisterEvent("ADDON_LOADED")

end

function Misspelled_OnEvent()
	if ( event == "VARIABLES_LOADED" ) then
		Misspelled_VariablesLoaded()
	elseif( event == "ADDON_LOADED" ) then
		Misspelled_CompatibilityCheck()
	end
end

local suggestionsDropdown

function Misspelled_VariablesLoaded()
	WordDict:Init()
	print("Misspelled: Dict Loaded")
	suggestionsDropdown = CreateFrame("Frame", "Test_DropDown", UIParent, "UIDropDownMenuTemplate");
	UIDropDownMenu_Initialize(suggestionsDropdown, MisspelledSuggestions_InitializeDropDown)
	--hooksecurefunc("ChatEdit_OnSpacePressed", Misspelled_OnSpacePressed)
	hooksecurefunc("ChatEdit_OnEscapePressed", Misspelled_OnEscapePressed)
	hooksecurefunc("ChatEdit_OnEnterPressed", Misspelled_OnEnterPressed)
	hooksecurefunc("ChatEdit_OnTextChanged", Misspelled_OnTextChanged)

	ChatFrameEditBox:SetScript("OnMouseUp", Misspelled_OnMouseUp)

	--Load user dict
	Misspelled:LoadUserDict()

	-- hooks for removing any misspelling highlightting in the text before it's sent to chat
	-- The Wow client will disconnect if you attempt to send a colored chat message.
	Misspelled_Orig_SendChatMessage = SendChatMessage;
	SendChatMessage = Misspelled_SendChatMessage;

end

function Misspelled_CompatibilityCheck()
	--print("SS: Addon_Loaded")
end

function Misspelled_OnTextChanged()
	--print ("TextChaged:", ChatFrameEditBox:GetCursorPosition())

	--track typing in the ChatEdit box, looking for word boundries.
	--spell check new words, tracking suggestions for misspelled words

	local text = ChatFrameEditBox:GetText()
	local pos = ChatFrameEditBox:GetCursorPosition()

	--If we currently just added a new char to the end of the line, see if it's a word boundry char
	if pos == #text then
		local lastChar = text:sub(-1, -1)
		if string.match(lastChar, "[ ,\.!\":]") ~= nil then
			Misspelled_OnSpacePressed()
		end
	end
end

function Misspelled_OnEscapePressed()

end

function Misspelled_OnEnterPressed()
	RightClickedWord = nil

	--before message is sent, remove and highlighting
	local newText = ChatFrameEditBox:GetText()
	newText = string.gsub(newText, "|cffff91c8", "")
	newText = string.gsub(newText, "|r", "")
	ChatFrameEditBox:SetText(newText)
end

function Misspelled_OnSpacePressed()
	local text = ChatFrameEditBox:GetText()
	--print("Space pressed:", text)

	--parse out just the words
	local words = {}
	local word
	local wordCount = 0

	for w in text:gmatch("[A-Za-z0-9_'À-ÿ]+") do
		wordCount = wordCount + 1
		words[wordCount] = w
	end

	if wordCount > 0 then
		word = words[wordCount]
		--ignore words all in uppercase
		if word == string.upper(word) then
			return
		end

		--ignore words with numbers in them
		if string.match(word, "[%d]") ~= nil then
			return
		end

		local correct = WordDict:Contains(word)
		--try the lower case version of the word
		if correct == false then
			correct = WordDict:Contains(string.lower(word))
		end
		if correct == false then
			local s = ""
			for i,v in ipairs(WordDict:Suggest(word)) do
				s = s .. v["Word"] .. ", "
			end
			print("Misspelled, try: (" .. s .. ")" )
		end
	end

	local h = Misspelled:HighlightMisspellings(text)
	--print("H:", h)
	if h ~= text then
		ChatFrameEditBox:SetText(h)
	end
end

function Misspelled_OnMouseUp(self, button)
	if button == "RightButton" then
		--check if we are positioned on a misspelled word
		local pos = ChatFrameEditBox:GetCursorPosition()

		for i, w in ipairs(WordLocations) do
			if	pos >= w.StartPos and pos <= w.EndPos then
				if WordCache[w.Word].Correct == false then
					RightClickedWord = w.Word
					RightClickedWordStartPos = w.StartPos
					RightClickedWordEndPos = w.EndPos
					--UIDropDownMenu_SetAnchor(MisspelledSuggestions_DropDown, 0, 0, "ChatFrameEditBox",
					--ToggleDropDownMenu(1, nul, MisspelledSuggestions_DropDown, "ChatFrameEditBox", 100, 180)
					ToggleDropDownMenu(1, nil, MisspelledSuggestions_DropDown, "cursor")
				else
					RightClickedWord = nil
				end
				break
			end
		end

		--print("cursor position:" , ChatFrameEditBox:GetCursorPosition())
	end
end

function Misspelled:OnSizeChanged(frame)

end

function MisspelledSuggestions_InitializeDropDown(level)
	if RightClickedWord == nil then return end
	if #RightClickedWord == 0 then return end

	info = UIDropDownMenu_CreateInfo()
	info.text = "Suggestions for: " .. RightClickedWord
	info.isTitle = 1
	UIDropDownMenu_AddButton(info)

	for i, s in ipairs(WordCache[RightClickedWord].Suggestions) do
		info = UIDropDownMenu_CreateInfo()
		info.owner = this:GetParent()
		info.text = s.Word
		info.isTitle = nil
		info.value = s.Word
		info.func = function() SuggestionsFrame_Click(s.Word) end
		--Add the above info to the optios menu as a button
		UIDropDownMenu_AddButton(info)
	end

	--Add a seperator
	info = UIDropDownMenu_CreateInfo()
	info.owner = this:GetParent()
	info.text = ""
	info.isTitle = nil
	info.value = ""
	info.notClickable = 1
	UIDropDownMenu_AddButton(info)

	info = UIDropDownMenu_CreateInfo()
	info.owner = this:GetParent()
	info.text = "Ignore All"
	info.isTitle = nil
	info.value = RightClickedWord
	info.func = function() SuggestionsFrame_Click("###IgnoreAll") end
	info.notClickable = nil
	UIDropDownMenu_AddButton(info)

	info = UIDropDownMenu_CreateInfo()
	info.owner = this:GetParent()
	info.text = "Add to Dictionary"
	info.isTitle = nil
	info.value = RightClickedWord
	info.func = function() SuggestionsFrame_Click("###AddToDictionary") end
	info.notClickable = nil
	UIDropDownMenu_AddButton(info)
end

function MisspelledSuggestions_DropDownOnLoad(self)
	UIDropDownMenu_Initialize(self, MisspelledSuggestions_InitializeDropDown)
end

function SuggestionsFrame_Click(value)
	--value should equal the suggestion clicked
	--print("Suggestion Clicked: ", value)

	local newText = ChatFrameEditBox:GetText()

	if value == "###IgnoreAll" then
		--print("Ignore:", RightClickedWord)
		WordCache[RightClickedWord].Correct = true
		WordCache[RightClickedWord].Suggestions = {}
	elseif value == "###AddToDictionary" then
		--print("Add to UserDict", RightClickedWord)
		Misspelled:AddToUserDict(RightClickedWord)
		WordCache[RightClickedWord].Correct = true
		WordCache[RightClickedWord].Suggestions = {}
	else
		--replace this word with the selected suggestion.
		newText = newText:sub(1, RightClickedWordStartPos - 1) .. value .. newText:sub(RightClickedWordEndPos +1 )
	end

	--remove any previous highlighting
	newText = string.gsub(newText, SPELLED_WRONG_HIGHLIGHT, "")
	newText = string.gsub(newText, "|r", "")

	ChatFrameEditBox:SetText(newText)

	RightClickedWord = nil

	Misspelled_OnSpacePressed()
end

function Misspelled:LoadUserDict()
	if Misspelled_DB == nil then
		Misspelled_DB = {UserDict = {}}
	end

	if Misspelled_DB.UserDict ~= nil then
		for k, v in pairs(Misspelled_DB.UserDict) do
			if WordDict.baseWords[k] == nil then
				WordDict.baseWords[k] = {v[1], v[2]}
			end
		end
	end
	print("Misspelled: User Dictionary Loaded")
end


function Misspelled:AddToUserDict(word)
	if word == nil then return end
	if #word == 0 then return end

	--Add the new word to the UserDict saved variable
	local pcode = WordDict:PhoneticCode(word)
	Misspelled_DB.UserDict[word] = {"", pcode}

	--And add it to the currently loaded dictionary
	WordDict.baseWords[word] = {"", pcode}

	--Fixup the WordCache
	WordCache[word].Correct = true
	WordCache[word].Suggestions = {}
end


function Misspelled:CheckLine(text)
	if text == nil then return end
	if #text == 0 then return end

	--March through the text, finding the words, and recording there start & end positions
	--Then see if the word is in the same location as last time we checked.
	--If it's a new word or changed word, spell check it.

	local ChatEditWords = {}
	local patt = "[A-Za-z0-9_'À-ÿ]+"
	local pos1
	local pos2
	pos1, pos2 = string.find(text, patt)
	pos = pos1

	local x = 1
	while pos1 ~= nil do
		pos = pos1
		local word = text:sub(pos1, pos2)
		WordLocations[x] = {["Word"] = word, ["StartPos"] = pos1, ["EndPos"] = pos2}


		if WordCache[word] == nil then
			local correct = false

			--Ignore words in all upper case
			if word == string.upper(word) then
				correct = true
			end

			--See if the dictionary contains the word
			if correct == false then
				correct = WordDict:Contains(word)
			end

			--Try the lower case version of the word
			if correct == false then
				correct = WordDict:Contains(string.lower(word))
			end

			WordCache[word] = {["Correct"] = correct, ["Suggestions"] = {}}
			if correct == false then
				local suggestions = {}
				suggestions = WordDict:Suggest(word)
				if #suggestions > 0 then
					WordCache[word].Suggestions = suggestions
				end
			end
		end
		x = x + 1
		pos1, pos2 = string.find(text, patt, pos2+1)
	end
end


--return a the
function Misspelled:HighlightMisspellings(text)
	if #text < 2 then return text end



	local newText = text
	--remove any previous highlighting
	newText = string.gsub(newText, SPELLED_WRONG_HIGHLIGHT, "")
	newText = string.gsub(newText, "|r", "")

	--print("newText", newText.."#")

	Misspelled:CheckLine(newText)

	--March backwords through the input text, highlighting misspellings
	--tprint(WordLocations)
	for x = #WordLocations, 1, -1 do
		local w = WordLocations[x]
		if WordCache[w.Word].Correct == false then
			--Insert highlighting
			newText = newText:sub(1, w.StartPos -1) .. SPELLED_WRONG_HIGHLIGHT .. newText:sub(w.StartPos, w.EndPos) .. FONT_COLOR_CODE_CLOSE .. newText:sub(w.EndPos+1)
		end
	end

	--Adjust the word StartPos and EndPos, whenever we added highlighting
	-- so the right click handler can detect the word right clicked.
	--March forward this time
	local n = 0 --NewCharsAddedCounter
	for x = 1, #WordLocations do
		local w = WordLocations[x]
		if WordCache[w.Word].Correct == false then
			w.StartPos = w.StartPos + n + #SPELLED_WRONG_HIGHLIGHT
			n = n + #SPELLED_WRONG_HIGHLIGHT
			w.EndPos = w.EndPos + n
			n = n + #FONT_COLOR_CODE_CLOSE
		end
	end

	return newText
end


--[[
--testing
require("bit")
require("WordDict")
require("Dict\\Dic_en_US")
Misspelled_DB = {}
WordDict:Init()
ChatFrameEditBox = {}
function ChatFrameEditBox:GetText()
	return "Leatherworking"
end

	--Load user dict
	if Misspelled_DB == nil then
		Misspelled_DB = {UserDict={}}
	elseif Misspelled_DB.UserDict == nil then
		Misspelled_DB.UserDict={}
	else
		LoadUserDict()
	end

Misspelled_DB = {UserDict = {}}
Misspelled_DB.UserDict["Test"] = {"aa", "PH"}
Misspelled:tprint(Misspelled_DB.UserDict["Test"])
--AddToUserDict("LeatherWorking")
--Misspelled:tprint(Misspelled_DB)
--Misspelled:tprint(WordDict.baseWords["LeatherWorking"])
---]]
